/*
 * Copyright (C) 2001 FUJITSU LABORATRIES LTD.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef _NETINET6_XCAST6_H_
#define _NETINET6_XCAST6_H_

#include <sys/types.h>
#include <netinet/ip6.h>

#define XCAST6_DESTOPT		/* Support for XCAST6 destination option */

#define	IPV6_RTHDR_TYPE_XCAST6		0x11	/* XXX  needs to register */
#define	IP6OPT_TYPE_XCAST6		0x27	/* XXX  needs to register */

/* XXX  this address is not registered... */
#define IN6ADDR_ALL_XCAST6_ROUTERS_INIT	\
	{{{ 0xff, 0x05, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, \
	    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10 }}}


/* XXX  Actually 0x7f, but limitation from the mbuf size */
#define XCAST6_MAX_DESTS		0x7e

/* X2U support level */
#define XCAST6_X2U_NEVER		0	/* no X2U support. */
#define XCAST6_X2U_SINGLEDEST		1	/* single dest. packet only. */
#define XCAST6_X2U_PREMATURE		2	/* every packet. N/A yet. */

/* Xcast6 Routing header */
typedef struct xcast6_rthdr {
	u_int8_t  x6r_nxt;	/* next header */
	u_int8_t  x6r_len;	/* length in units of 8 octets */
	u_int8_t  x6r_type;	/* IPV6_RTHDR_TYPE_XCAST */
	u_int8_t  x6r_reserved;	/* reserved */
	u_int8_t  x6r_flags;	/* version, anonymity, Xcast, DSCP */
	u_int8_t  x6r_ndest;	/* # of destinations */
	u_int16_t x6r_cksum;	/* checksum; 1's complement */
	u_int8_t  x6r_chid[16];	/* channel identifier */
	/* followed by a bitmap and destionation addresses */
} xcast6_rthdr_t;

#define	XCAST6_VERSION			1
#define	XCAST6_RTHDR_VERSION_MASK	0xf0
#define	XCAST6_RTHDR_VERSION_SHIFT	4
#define XCAST6_RTHDR_ANON		0x08
#define XCAST6_RTHDR_DONTX2U		0x04
#define XCAST6_RTHDR_DSCP		0x02

/* The size of a bitmap in routing header. */
#define	XCAST6_BM_SIZE(ndest)	(((((ndest) + 7) / 8) + 7) & ~7)
#define XCAST6_RTHDRLEN(ndest)		\
	(sizeof(xcast6_rthdr_t) +	\
	 XCAST6_BM_SIZE(ndest) +	\
	 sizeof(struct in6_addr) * (ndest))

/* 
 * Xcast6 semi-permeable tunneling.
 * Use hop-by-hop extension header for tunneling.
 */

/* Xcast6 Hop-by-Hop options header */
typedef struct xcast6_hbh {
	u_int8_t  x6h_type;		/* IP6OPT_TYPE_XCAST6 */
	u_int8_t  x6h_optdlen;		/* >= 0 */
	/* Xcast hop-by-hop extensions follow... */
} xcast6_hbh_t;

typedef struct xcast6_hbhso {		/* Xcast6 suboption in hbh */
	u_int8_t  x6hs_type;
	u_int8_t  x6hs_optdlen;		/* >= 0 */
} xcast6_hbh_so_t;

#define XCAST6_HBH_OPTLEN	(sizeof(xcast6_hbh_t) - 2)
#define XCAST6_HBH_HDRLEN		\
	((sizeof(struct ip6_hbh) + sizeof(xcast6_hbh_t) + 7) & ~7)

#define XCAST6_SUBOPT_ICMP6	0x01

typedef struct xcast6_hbhso_icmp6 {
	u_int8_t x6hi_type;
	u_int8_t x6hi_optdlen;
	u_int8_t x6hi_nth;		/* n-th destination */
} xcast6_hbhso_icmp6_t;

/* Xcast6 destination options header */
typedef struct xcast6_dest {
	u_int8_t  x6d_type;		/* IP6OPT_TYPE_XCAST6 */
	u_int8_t  x6d_optdlen;		/* >= 0 */
	/* Xcast destination ports follow... */
} xcast6_dest_t;

#define XCAST6_DEST_HDRLEN(ndest)		\
	((sizeof(struct ip6_dest) +		\
	  sizeof(xcast6_dest_t) +		\
	  sizeof(u_int16_t) * (ndest) + 7) & ~7)

/* Sysctl objects. Should be in route6.h? */
#define ROUTE6CTL_XCAST6_ENABLE		1
#define ROUTE6CTL_XCAST6_SPTUNNEL	2
#define ROUTE6CTL_XCAST6_X2ULEVEL	3
#define ROUTE6CTL_XCAST6_ALLOWSITELOCAL	4
#define ROUTE6CTL_MAXID			5

#define ROUTE6CTL_NAMES { \
	{ 0, 0 }, \
	{ "xcast6_enable", CTLTYPE_INT }, \
	{ "xcast6_sptunnel", CTLTYPE_INT }, \
	{ "xcast6_x2ulevel", CTLTYPE_INT }, \
	{ "xcast6_allowsitelocal", CTLTYPE_INT }, \
}

#ifdef _KERNEL
/* *_var part.  better create xcast6_var.h? */

/* Bitmap structure internally used in XCAST6 routing. */
typedef struct {
	u_int8_t		b[XCAST6_BM_SIZE(XCAST6_MAX_DESTS)];
} xcast6_bm_t;

/* Routing table structure. */
typedef struct {
	struct route_in6	x6_rt;
	/* Will be added members when needed. */
} xcast6_routetab_t;

/*
 * Xcast6 auxiliary structure used for Xcast6 options.
 * This structure is put as mbuf_aux data.
 */
typedef struct xcast6_ctl {
	u_int16_t   x6c_port;		/* destination port number */
	int         x6c_ndest;		/* number of destinations */
	int         x6c_nth;		/* nth destination (port option)  */
	int         x6c_icmp6nth;	/* icmp6 nth destination  */
	int         x6c_reflectable;	/* icmp6 reflectable? */
	int	    x6c_hbhseen;	/* seen XCAST6 hbh before inner IP6 */
	xcast6_routetab_t *x6c_rtabspace; /* routing table space. */
	size_t	    x6c_rtablen;	/* length that can be used for rtab. */
	u_int64_t   _pad;
	/* more options may be added... */
} xcast6_ctl_t;

extern int xcast6_enable;
extern int xcast6_sptunnel;
extern int xcast6_x2ulevel;
extern int xcast6_allowsitelocal;

int xcast6_branch __P((struct mbuf *, xcast6_rthdr_t *, int *, int));
void xcast6_launch __P((struct mbuf *, int *, int));
xcast6_ctl_t *xcast6_addctl __P((struct mbuf *));
xcast6_ctl_t *xcast6_getctl __P((struct mbuf *));
void xcast6_delctl __P((struct mbuf *));
int xcast6_isreflectable __P((struct mbuf *, int));
int xcast6_process_hopopts __P((struct mbuf *, u_int8_t *, int *));
int xcast6_process_destopts __P((struct mbuf *, u_int8_t *, int *));
int xcast6_hasdestopt __P((struct ip6_pktopts *));
int xcast6_is_addr_xcast6 __P((struct in6_addr *));

/* Should be in route6.h? */
int route6_sysctl __P((int *, u_int, void *, size_t *, void *, size_t));
#endif

#endif /* ! _NETINET6_XCAST6_H_ */
